import json,os
from pprint import *
import requests
from web3 import Web3
from eth_account import Account
from web3.gas_strategies.time_based import construct_time_based_gas_price_strategy
from web3.middleware import geth_poa_middleware

class eth_chain_client:
    def __init__(self,
                 network_url="https://eth-ropsten.alchemyapi.io/v2/zxbIGpQIr15V6wLhHgHVJir3BaIctTGX",
                 poa=False,
                 keystore_dir='eth_wallet/conf/keystores/',
                 keystore_filename='root.keystore',
                 pwd='123'):
        """
        eth_chain_client

        :param network_url: sample: \n
        https://rinkeby.infura.io/v3/cf9f29fb4f674714be8047085920bbd9
        """
        self.poa = poa
        self.keystore_dir = keystore_dir
        self.keystore_filename = keystore_filename
        self.pwd = pwd 
        self.network_url = network_url
        
        if not os.path.exists(self.keystore_dir+self.keystore_filename):
            self.init_keystore_file()

        with open(self.keystore_dir+self.keystore_filename,"r") as r:
            self._privatekey = Account.decrypt(json.loads(r.read()),self.pwd).hex()

        self._address = Account.privateKeyToAccount(self._privatekey).address
        
        self.web3 = Web3(Web3.HTTPProvider(self.network_url))
        
        if self.poa:
            self.web3.middleware_onion.inject(geth_poa_middleware, layer=0)
    
    @property
    def private_key(self):
        return self._privatekey

    @property
    def address(self):
        return self._address
    
    def update_network(self,url):
        self.network_url = url
        return True

    def create_wallet(self,pwd):
        with open(self.keystore_dir+self.keystore_filename,"w") as w:
            acc = Account.create()
            address = acc.address
            private = acc.privateKey 
            self.address = address
            self._privatekey = private
            w.write(json.dumps(Account.encrypt(private,pwd)))
            return {address:private.hex()}

    def recover_wallet(self,private,pwd):
        with open(self.keystore_dir+self.keystore_filename,"w") as w:
            acc = Account.privateKeyToAccount(private)
            address = acc.address
            self._address = address
            self._privatekey = private
            w.write(json.dumps(Account.encrypt(private,pwd)))
            return {address:private}
        
    def get_wallet_from_file(self,keystore_file):
        with open(keystore_file,"r") as r:
            private = Account.decrypt(json.loads(r.read()),self.pwd).hex()
            address = Account.privateKeyToAccount(private).address
            return {address:private}

    def init_keystore_file(self):
        """
        keystore初始化
        """
        with open(self.keystore_dir+self.keystore_filename,"w") as w:
            w.write(json.dumps(Account.encrypt(Account.create().privateKey,self.pwd)))
    
    def sign_and_sendRaw(self,params, private_key):
        """
        签名发送函数

        :param params: _description_
        :param private_key: _description_
        :return: _description_
        """
        # 使用私钥签名
        signed_tx = self.web3.eth.account.signTransaction(params, private_key=private_key)
        # 发送交易，并获得交易hash
        tx_hash = self.web3.eth.sendRawTransaction(signed_tx.rawTransaction)
        # 返回交易hash
        print("tx_hash: {}".format(tx_hash.hex))
        return tx_hash.hex()
    
    def get_balance(self,address):
        """
        获取余额

        :param address: _description_
        :return: _description_
        """
        return float(self.web3.fromWei(self.web3.eth.get_balance(address),"Ether"))

    def transfer_eth(self,target_address, amount,private_key):
        """
        以太坊转账

        :param target_address: _description_
        :param amount: _description_
        :param private_key: _description_
        :return: _description_
        """
        address = self.web3.eth.account.from_key(private_key).address
        # 获取用户的nonce
        nonce = self.get_nonce(address)
        # 构建交易数据包
        params = {
            'nonce': nonce,
            'to': target_address,
            'value': self.web3.toWei(amount, 'ether'),
            'gas': 210000,
            'gasPrice': self.web3.toWei('150', 'gwei'),
            # 'gasPrice': self.web3.toHex(self.web3.toWei(90, "Gwei")),
            'from': address,
        }
        return self.sign_and_sendRaw(params,private_key)

    def contract_fun(self,contractAddress,abi_file,private_key,function_name,args=None,call=False):
        """
        contractAddress = 合约地址 \n
        contract_abi = 合约abi \n
        private_key = 你的私钥  \n
        args = 合约参数,tuple格式 \n
        不用输入合约地址 TODO
        """
        # 这个只是hardhat生成的abi，其他的可以做个适配
        with open(abi_file,"r") as r:
            contract_abi = json.loads(r.read())
        address = self.web3.eth.account.from_key(private_key).address
        contractAddress = self.web3.toChecksumAddress(contractAddress)
        
        # 通过地址找到nonce
        nonce = self.get_nonce(address)
        
        # 定义合约对象
        contract = self.web3.eth.contract(address=contractAddress, abi=contract_abi)
        params = {'from': address,'nonce':nonce}
        
        if not call:
                print("contract.functions.{}{}".format(function_name,args))
                txn = eval("contract.functions.{}{}".format(function_name,args))
                self.web3.eth.set_gas_price_strategy(construct_time_based_gas_price_strategy(max_wait_seconds=60,sample_size=10))
                gasPrice = self.web3.eth.generateGasPrice()
                estimateGas = txn.estimateGas(params)
                params.update({
                    'gas': int(estimateGas*1.1),
                    'gasPrice': int(gasPrice*1.5)
                })
                txn = txn.buildTransaction(params)
        elif args!=None:
            return eval("contract.functions.{}().call()".format(function_name))
        else:
            return eval("contract.functions.{}(\"{}\").call()".format(function_name,args))
        receipt = self.sign_and_sendRaw(txn,private_key)
        return self.web3.eth.wait_for_transaction_receipt(timeout=360,transaction_hash=receipt)

    def get_gas_price(self):
        headers = {
            "user-agent": "Mozilla/5.0 (Linux; Android 6.0; Nexus 5 Build/MRA58N) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/100.0.4896.127 Mobile Safari/537.36"
            }
        url = 'https://api-rinkeby.etherscan.io/api?module=proxy&action=eth_gasPrice&apikey=JKXR76EA1NK9XSKES1T2BNYBZ41T3BJ2JH'
        resp_data = json.loads(requests.get(url,headers=headers).text)
        return int(resp_data['result'],16)
    
    def get_nonce(self,address):
        nonce = self.web3.eth.getTransactionCount(address)
        return nonce


if __name__ == '__main__':

    block_chain_demo1 = eth_chain_client()
    
    to_address = "0x7841dEf061390f68c124B04490201C9279da2f02"
    private_key = "0x808022e3767e57091c898c494a68ba628a6a3140c2219913aef0dc947e0959c1"
    # tx_hash = block_chain_demo1.transfer_eth(to_address,10000,private_key)
    # tx_hash = block_chain_demo1.deploy_contract("./test.sol",private_key)
    # tx_hash = block_chain_demo1.contract_fun("0x75c2cc125bafde7f9ab9b349b237057a6c40e0ee","test.abi",private_key,"get")
    # tx_hash = block_chain_demo1.contract_fun("0x75c2cc125bafde7f9ab9b349b237057a6c40e0ee","test.abi",private_key,"set",44)
    
    block_chain_demo1.get_nonce("0xf50b35bb81d9e089a5b96971e9bd8ba37802bb89")